#!/usr/bin/env python

import os
import platform

from setuptools import Extension, setup

work_dir = os.path.dirname(os.path.realpath(__file__))
mod_dir = os.path.join(work_dir, 'src', 'confluent_kafka')
ext_dir = os.path.join(mod_dir, 'src')

# On Un*x the library is linked as -lrdkafka,
# while on windows we need the full librdkafka name.
if platform.system() == 'Windows':
    librdkafka_libname = 'librdkafka'
else:
    librdkafka_libname = 'rdkafka'

module = Extension(
    'confluent_kafka.cimpl',
    libraries=[librdkafka_libname],
    sources=[
        os.path.join(ext_dir, 'confluent_kafka.c'),
        os.path.join(ext_dir, 'Producer.c'),
        os.path.join(ext_dir, 'Consumer.c'),
        os.path.join(ext_dir, 'Metadata.c'),
        os.path.join(ext_dir, 'AdminTypes.c'),
        os.path.join(ext_dir, 'Admin.c'),
    ],
)

setup(ext_modules=[module])
